/*
This tap:
    - captures the broadcast time offset table
    - overrides the firmware utc_to_localtime() function

The firmware will convert the timestamps on EPG events based on the current local
time offset, rather than the time offset that will be in effect when the EPG
event actually occurs.  This tap modifies the firmware conversion function,
using information that is broadcast over the air.

To be effective,
- you must have the toppy time set to auto.
- the broadcaster must be sending accurate information.
- you should add this to autostart, since any EPG events captured before it has
been started will be convert in the original way.
- it will only have any effect during the week before a time offset (daylight
saving time) change.


Usage:
    run - preferably from autostart
    menu via Recall key, or if tsrcommander loaded, use Exit

    v0.01 - initial
*/
typedef unsigned char  UCHAR;
typedef unsigned short USHORT;
typedef unsigned long  ULONG;
#include <tap.h>
#include "exTap.h"
#include "tf5800.h"

#define ID_USERTAP 0xbdb01901
TAP_ID( ID_USERTAP );

TAP_PROGRAM_NAME("tot v0.01");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("a tap to convert from utc to local time properly");
TAP_ETCINFO(__DATE__);

//----------------------------------------------------------------------------
//
//tsr_commander stuff
//
#include "TSRCommander.inc"

//----------------------------------------------------------------------------
#define is_firmware(p) (((p&0xfc000000)==0x80000000) ? 1 : 0)

extern type_menu main_menu;
dword fw_DbgSi_LoadDmx;

int tot_enabled = 0;
int utctolocal_time_enabled = 0;

int utctolocal_time_hook_id;
int tot_hook_id;

typedef struct {
    dword next_change;
    int offset_polarity;
    int current_offset;
    int next_offset;
    byte country_code[4];
    int country_region_id;
}type_tot;

type_tot tot = {0};

//----------------------------------------------------------------------------
//
//main
//

extern void TSRCommanderConfigDialog();
extern bool TSRCommanderExitTAP();
extern int is_tsr_commander_running(void);

extern int TAP_Main(void);
extern dword TAP_EventHandler(word event, dword param1, dword param2);
extern int main_menu_handler(type_menu *menu, dword param1);
extern int key_handler(dword param1, dword param2);

extern int apply_utctolocal_time_hook(void);
extern int remove_utctolocal_time_hook(void);
extern dword utctolocal_time_hook(dword param1, dword param2);

extern int apply_tot_hook(void);
extern int remove_tot_hook(void);
extern dword tot_hook(dword param1, dword param2);

extern int init(void);
extern int close(void);
dword utctolocal_time(dword utc);
int display_time_offset();

//----------------------------------------------------------------------------
//
//main
//
char *tap_name;
type_menu main_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {3, "xxx",           },
        {0, "Display Time Offset",  },
        {1, "Disable TOT",   },
        {2, "Exit Tap",      },
    },
};

//----------------------------------------------------------------------------
//
//tsr_commander stuff
//
bool TSRCommanderExitTAP()
{
    close();
    TAP_Exit();
    return FALSE;
}

void TSRCommanderConfigDialog()
{
    TAP_ExitNormal();
    ShowMenu(&main_menu);
}

int is_tsr_commander_running(void)
{
    int i;
	type_TAPProcess *TAPProcess;

    for (i=0; i<16; i++) {
        TAPProcess = (type_TAPProcess *)(fw_tap_table + (i*32));
        if (TAPProcess->firmwaregp) {
            if (TAPProcess->header) {
                if (TAPProcess->header->id == 0x807a10b2) {  //tsr_commander
                    return 1;
                }
                if (TAPProcess->header->id == 0x81001021) {  //tap_commander
                    return 1;
                }
            }
        }
    }
    return 0;
}

//----------------------------------------------------------------------------
int TAP_Main(void)
{
    int ret;

    tap_name =  (char *)__tap_program_name__;
    sprintf(strLog, "Opening %s", tap_name);
    log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_SHORT);

    //determine support for different firmwares
    ret = exTAP_Init();
    if (ret) {
        sprintf(strLog, "Unsupported platform - %d:%04x", fw_sysid, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
        return 0;
    }

    if ((fw_sysid==456)&&(_appl_version==0x1225)) {
        fw_DbgSi_LoadDmx = 0x8007c678;
    }
    else if ((fw_sysid==456)&&(_appl_version==0x1288)) {
        fw_DbgSi_LoadDmx = 0x8007c6ec;
    }
    else if ((fw_sysid==456)&&(_appl_version==0x1326)) {
        fw_DbgSi_LoadDmx = 0x8007df58;
    }
    else {
        sprintf(strLog, "Unsupported platform - %d:%04x", fw_sysid, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
        return 0;
    }

    init();
    return 1;                 //0=application mode, 1=tsr mode
}

dword TAP_EventHandler(word event, dword param1, dword param2)
{
    ULONG state, subState;
    int ret;

    switch (event) {
        case EVT_IDLE:
            if (message_showing) {
                if (TAP_GetTick() > message_showing) {
                    TAP_Osd_Delete(message_rgn);    // release rgn-handle
                    message_showing = 0;
                    message_rgn = 0;
                }
            }

    		TSRCommanderWork();
            break;

        case EVT_KEY:
            TAP_GetState(&state, &subState);
            if (main_menu.visible) {
                param1 = main_menu_handler(&main_menu, param1);
            }
            else {
                if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                    param1 = key_handler(param1, param2);
                }
            }
	        break;
        case EVT_HOOK:
            if (param1 == tot_hook_id) param1 = tot_hook(param1, param2);
            if (param1 == utctolocal_time_hook_id) param1 = utctolocal_time_hook(param1, param2);
            break;
        default:
            break;
    }
    return param1;
}


int main_menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int chNum;
    int nTvSvc, nRadioSvc;
    int tv_aspect;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
            TAP_EnterNormal();
			switch (selection) {
				case 0:
				    display_time_offset();
				    return 0;
				case 1:
                    if (tot_enabled) {
                        sprintf(strLog, "Disabling TOT");
                        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                        strcpy(menu->item[2].str, "Enable TOT");
                        remove_tot_hook();
                        remove_utctolocal_time_hook();
                    }
                    else {
                        sprintf(strLog, "Enabling TOT");
                        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                        apply_tot_hook();
                        apply_utctolocal_time_hook();
                        if (tot_enabled && utctolocal_time_enabled) {
                            strcpy(menu->item[2].str, "Disable TOT");
                        }
                        else {
                            remove_tot_hook();
                            remove_utctolocal_time_hook();
                            strcpy(main_menu.item[2].str, "Enable TOT");
                        }
                    }
				    return 0;
				case 2:
				    close();
				    TAP_Exit();
				    return 0;
			}
            break;
        case RKEY_Recall:
			DeleteMenu(menu);
            TAP_EnterNormal();
            return param1; // cascades key to other TAP's
        case RKEY_Exit:
			DeleteMenu(menu);
            TAP_EnterNormal();
            return 0; // exit without Cascading
        default:
	        TAP_Win_Action(&menu->win, param1); // send all other key's to menu-Win
    		return 0;
    }
    return param1;
}


int key_handler(dword param1, dword param2)
{
    switch(param1) {
        //only if tsrcommander is not running
        case RKEY_Recall:
            if (!is_tsr_commander_running()) {
                TAP_ExitNormal();
                ShowMenu(&main_menu);
                return 0;
            }
            break;
    }
    return param1;
}

//-----------------------------------------------------------------------------
//
//hook the utctolocal_time() function
//

type_hook utctolocal_time_hook_patch[] = {
// sysid, firmware, enable, address,    length, original_data
    {456, 0x1225,   1,      0x8012c34c, 8,      {0x27, 0xbd, 0xff, 0xd0,  0xaf, 0xbf, 0x00, 0x28 }, },  //fw_utctolocal_time()
    {456, 0x1288,   1,      0x8012fb08, 8,      {0x27, 0xbd, 0xff, 0xd0,  0xaf, 0xbf, 0x00, 0x28 }, },  //fw_utctolocal_time()
    {456, 0x1326,   1,      0x80131598, 8,      {0x27, 0xbd, 0xff, 0xd0,  0xaf, 0xbf, 0x00, 0x28 }, },  //fw_utctolocal_time()
};

int apply_utctolocal_time_hook(void)
{
    utctolocal_time_hook_id = exTAP_ApplyHook(&utctolocal_time_hook_patch[0]);
    if (utctolocal_time_hook_id < 0) {
        utctolocal_time_enabled = 0;
        TAP_Print("\nERROR: apply_utctolocal_time_hook(), firmware version %04x not supported", _appl_version);
        return -1;
    }
    utctolocal_time_enabled = 1;
    return 0;
}

int remove_utctolocal_time_hook(void)
{
    int ret;

    if (utctolocal_time_hook_id >=0) {
        ret = exTAP_RemoveHook(utctolocal_time_hook_id);
        if (ret == 0) utctolocal_time_hook_id = -1;
    }
    utctolocal_time_enabled = 0;
    return 0;
}

//
//typedef struct {
//    word    mjd;
//    byte    hour;
//    byte    minute;
//}type_taptime;
//
// type_taptime fw_utctolocal_time(type_taptime utc);
//
dword utctolocal_time_hook(dword param1, dword param2)
{
    type_hook_registers *reg;
    dword utc;
    dword local;

//TAP_Print("\nEVT_HOOK_utctolocal_time, param1=%x, param2=%x", param1, param2);
    reg = (type_hook_registers *)param2;
    utc = (dword)(reg->a0);
//TAP_Print("\nutctolocal_time(a0=%x)", reg->a0);

    local = utctolocal_time(utc);
    if (local) {
        reg->v0 = (dword)local;
        return -1;          //bypass the firmware routine
    }

    return param1;          //use the firmware routine
}

//----------------------------------------------------------------------------
//
//hook the tot() function
//

type_hook tot_hook_patch[] = {
// sysid, firmware, enable, address,    length, original_data
    {456, 0x1225,   1,      0X8010f5a4, 8,      {0x27, 0xbd, 0xff, 0xc8,  0xaf, 0xbf, 0x00, 0x2c }, },  //fw_process_tot()
    {456, 0x1288,   1,      0x8010fdd0, 8,      {0x27, 0xbd, 0xff, 0xc8,  0xaf, 0xbf, 0x00, 0x2c }, },  //fw_process_tot()
    {456, 0x1326,   1,      0x801116d0, 8,      {0x27, 0xbd, 0xff, 0xc8,  0xaf, 0xbf, 0x00, 0x2c }, },  //fw_process_tot()
};

int apply_tot_hook(void)
{
    tot_hook_id = exTAP_ApplyHook(&tot_hook_patch[0]);
    if (tot_hook_id < 0) {
        tot_enabled = 0;
        TAP_Print("\nERROR: apply_tot_hook(), firmware version %04x not supported", _appl_version);
        return -1;
    }
    tot_enabled = 1;
    return 0;
}

int remove_tot_hook(void)
{
    int ret;

    if (tot_hook_id >=0) {
        ret = exTAP_RemoveHook(tot_hook_id);
        if (ret == 0) tot_hook_id = -1;
    }
    tot_enabled = 0;
    return 0;
}

//typedef struct
//{
//    UCHAR  table_id                         : 8;
//    UCHAR  section_syntax_indicator         : 1;
//    UCHAR  reserved_future_use              : 1;
//    UCHAR  reserved1                        : 2;
//    USHORT section_length                   :12;
//    UCHAR  UTC_time[5];                   //:40;
//    UCHAR  reserved2                        : 4;
//    USHORT descriptors_loop_length          :12;
//
//    UCHAR  num_descriptors;
//    descriptor_type *descriptor;
//
//    ULONG  CRC_32                           :32;
//} time_offset_section_type;

//typedef struct
//{
//    UCHAR  descriptor_tag                       : 8;      //0x58
//    UCHAR  descriptor_length                    : 8;
//
//    UCHAR  *country_code[3];                  //:24;
//    UCHAR  *country_region_id;                //: 6;
//    UCHAR  *reserved;                         //: 1;
//    UCHAR  *local_time_offset_polarity;       //: 1;
//    USHORT *local_time_offset;                //:16;
//    UCHAR  *time_of_change[5];                //:40;
//    USHORT *next_time_offset;                 //:16;
//} local_time_offset_descriptor_type;

dword tot_hook(dword param1, dword param2)
{
    type_hook_registers *reg;
    dword a0;
    byte a1;
    byte a2;
    dword *buf;
    dword ret;
    int tot_desc;
    byte *tos;
    int local_time_offset_polarity;


//TAP_Print("\nEVT_HOOK_tot, param1=%x, param2=%x", param1, param2);
    reg = (type_hook_registers *)param2;
    a0 = (dword)(reg->a0);
//    dump_reg(reg);
//    TAP_Print("\na0 = %x", a0);

    //DbgSi_LoadDmx(a0, &a1, (byte *)&a2, &buffer);
    ret = exTAP_CallFirmware((void *)fw_DbgSi_LoadDmx, a0, (dword)&a1, (dword)&a2, (dword)&buf, 0,0,0,0,0,0,0);
//    TAP_Print("\nret = %x", ret);
//    TAP_Print("\na1 = %x", a1);
//    TAP_Print("\na2 = %x", a2);
//    TAP_Print("\nbuf = %x", buf);
    if (is_firmware(*buf)) {
        tos = (byte *)*buf;
//        dump((dword)tos, 0x40);
        tot_desc = 0xa;
        //should really scan for multiple descriptors
        if (tos[tot_desc] == 0x58) {
            tot.country_code[0] = tos[tot_desc+0x2];
            tot.country_code[1] = tos[tot_desc+0x3];
            tot.country_code[2] = tos[tot_desc+0x4];
            tot.country_code[3] = '\0';
            tot.country_region_id = tos[tot_desc+0x5] >> 2;

            tot.offset_polarity = tos[tot_desc+0x5] & 1;

            tot.current_offset = tos[tot_desc+0x6];
            tot.current_offset <<= 8;
            tot.current_offset += tos[tot_desc+0x7];

            tot.next_change = tos[tot_desc+0x8];
            tot.next_change <<= 8;
            tot.next_change += tos[tot_desc+0x9];
            tot.next_change <<= 8;
            tot.next_change += tos[tot_desc+0xa];
            tot.next_change <<= 8;
            tot.next_change += tos[tot_desc+0xb];

            tot.next_offset = tos[tot_desc+0xd];
            tot.next_offset <<= 8;
            tot.next_offset += tos[tot_desc+0xe];
        }
    }

    return -1;      //bypass the firmware routine
}


//-----------------------------------------------------------------------------
int init(void)
{
    TAP_Print("\nOpening %s", tap_name);
    TAP_Print("\n_tap_startAddr=%x", _tap_startAddr);

//    exTAP_Key_Extend2(1);

    main_menu.visible = 0;

    apply_tot_hook();
    apply_utctolocal_time_hook();
    if (tot_enabled && utctolocal_time_enabled) {
        strcpy(main_menu.item[2].str, "Disable TOT");
    }
    else {
        remove_tot_hook();
        remove_utctolocal_time_hook();
        strcpy(main_menu.item[2].str, "Enable TOT");
    }

    TSRCommanderInit(0, TRUE);
    return 0;
}

int close(void)
{
    DeleteMenu(&main_menu);
    remove_tot_hook();
    remove_utctolocal_time_hook();

    sprintf(strLog, "Closing  %s", tap_name);
    log_message(LOG_SERIAL|LOG_SCREEN_WAIT, strLog, DELAY_SHORT);
    if (OpLog) TAP_Hdd_Fclose(OpLog);

//    exTAP_Key_Extend2(0);        //undo Key Mapping patch
    TSRCommanderErase();
    TAP_Print("\nClosing %s", tap_name);
    return 0;
}

//----------------------------------------------------------------------------
//For what it's worth, the daylight savings rule in the UK is: between 1.00am on
//the last Sunday in March and 1.00am on the last Sunday in October, British
//Summer Time is in force; outside those times, GMT is in force.

dword utctolocal_time(dword utc)
{
    dword local;

    //is tot hook running?
    if (tot.next_change > 0) {
        if (utc < tot.next_change) {
//catch day overrflow
//check if offset is negative
            if (tot.offset_polarity) {
                local = utc - tot.current_offset;
                if ((local & 0xffff) > 0x173b) {  //23:59
                    local = local + 0x1800;
                    local = local - 0x00010000;
                }
            }
            else {
                local = utc + tot.current_offset;
                if ((local & 0xffff) > 0x173b) {  //23:59
                    local = local - 0x1800;
                    local = local + 0x00010000;
                }
            }
        }
        else {
            if (tot.offset_polarity) {
                local = utc - tot.next_offset;
                if ((local & 0xffff) > 0x173b) {
                    local = local + 0x1800;
                    local = local - 0x00010000;
                }
            }
            else {
                local = utc + tot.next_offset;
                if ((local & 0xffff) > 0x173b) {
                    local = local - 0x1800;
                    local = local + 0x00010000;
                }
            }
        }
    }
    else {
        //use the built in method ...
        return 0;
    }

    return local;
}

int display_time_offset()
{
    word year;
    byte month, day, weekDay;
    char str[0x200];

    if (tot.next_change) {
        TAP_ExtractMjd((tot.next_change>>16), &year, &month, &day, &weekDay);
        TAP_Print("\nCOUNTRY CODE        :   %s", tot.country_code);
        TAP_Print("\nCOUNTRY REGION ID   :   %x", tot.country_region_id);
        TAP_Print("\nOFFSET POLARITY     :   %x", tot.offset_polarity);
        TAP_Print("\nCURRENT OFFSET      :   %02d:%02d", tot.current_offset>>8, tot.current_offset&0xff);
        TAP_Print("\nNEXT TIME CHANGE DUE:   %02d:%02d %02d-%02d-%04d", (tot.next_change>>8)&0xff, tot.next_change&0xff, day,  month, year);
        TAP_Print("\nNEXT OFFSET         :   %02d:%02d", tot.next_offset>>8, tot.next_offset&0xff);

        TAP_Print("\n");

        sprintf(str, "COUNTRY CODE  :  %s\n"
                     "COUNTRY REGION ID  :  %x\n"
                     "CURRENT OFFSET  :  %c%02d:%02d\n"
                     "NEXT TIME CHANGE DUE  :  %02d:%02d %02d-%02d-%04d\n"
                     "NEXT OFFSET  :  %c%02d:%02d",
                     tot.country_code,
                     tot.country_region_id,
                     (tot.offset_polarity)?'-':'+', tot.current_offset>>8, tot.current_offset&0xff,
                     (tot.next_change>>8)&0xff, tot.next_change&0xff, day,  month, year,
                     (tot.offset_polarity)?'-':'+', tot.next_offset>>8, tot.next_offset&0xff);
        ShowMessage2(str, DELAY_MEDIUM, 0);
    }
    return 0;
}
//----------------------------------------------------------------------------


